#include "gspinner.h"
/*
 * This file is released to the public domain, all responsibility for it is waived.
 * 
 * Original Author: Jeremy Latt (jlatt at faceprint dot com)
 * Current Maintainer: Pratik Dave (gspinner@pratikdave.com)
 *
 */
/* 
 * This is version 2.3 - fixed zombies when changing channels,
 *                       cleaned code up a bit more, added DEBUG,
 *                       made process more resilient to periodic
 *                       http/net hiccups
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif
#ifndef DEBUG
int DEBUG = 0;
#else 
int DEBUG = 1;
#endif

/* channel management */
GList *Genres;
Channel *current, *None;

/* display labels and the like */
GtkWidget *lartist; 
GtkWidget *lalbum;
GtkWidget *lsong;
GtkWidget *lchannel;
GtkWidget *lmenu;

/* trplayer */
pid_t trpid;
pthread_cond_t cond = PTHREAD_COND_INITIALIZER;

struct {
  char *raw;
  long pos;
  long length;
} data;



/*
 * CALLBACKS: menu_activate
 */
void menu_activate(GtkObject *o, Channel *c) {
  if (c && c != current) {
    if (current->id == 0) { 
      current = c;
      pthread_cond_signal(&cond);
      return;
    } else {
      current = c;
      kill(trpid, SIGKILL);
    }
  }
}

/*
 * UTILITY FUNCTIONS 
 */

FILE *fopen_home(char *path, char *perm) {
  char buf[256];
  snprintf(buf, 255, "%s/%s", getenv("HOME"), path);
  return fopen(buf, perm);
}

int System(char *command) {
  int pid, status;
  if (command == 0) return 1;
  pid = fork();
  if (pid == -1) return -1;
  if (pid == 0) {
    char *argv[4];
    argv[0] = "sh";
    argv[1] = "-c";
    argv[2] = command;
    argv[3] = 0;
    execve("/bin/sh", argv, environ);
    printf("I SHOULD NEVER BE HERE!!!!!!!!\n");
    exit(127);
  }
  trpid = pid;
  do {
    if (waitpid(pid, &status, 0) == -1) {
      trpid = 0;
      if (errno != EINTR) return -1;
    } else {
      // If they change channels on us we must remember
      // to detect the abnormal exit condition and terminate
      // the process to avoid zombies of trplayer
      if (!WIFEXITED(status)) {
	if (DEBUG) {
	  printf("Didn't return normal exit status, going after the process\n");
	}
	kill(pid, SIGKILL);
      }
      trpid = 0;
      return status;
    }
  } 
  while(1);
}

void finish(char *s) {
  char *t = s, *r =  NULL;
  while (t && (t = strstr(t, "&amp;"))) {
    r = strdup(t+5);
    strcpy(t+1, r);
    free(r);
  }
  t = s;
  while (t && (t = strstr(t, "&quot;"))) {
    r = strdup(t+6);
    *t = '"';
    strcpy(t+1, r);
    free(r);
  }
  /*t = s;
    while (t && (t = strchr(t, ':'))) *t = '-';*/
  s = realloc(s, strlen(s)+1);
}

/* 
 * PSEUDO-CLASSES: Channel, Genre
 */

Channel *Channel_new(char *name, int id) {
  Channel *c = malloc(sizeof(Channel));
  c->name = strdup(name);
  c->id = id;
  c->item = gtk_menu_item_new_with_label(c->name);
  gtk_signal_connect(GTK_OBJECT(c->item), "activate", 
		     GTK_SIGNAL_FUNC(menu_activate), c);
  gtk_widget_show(c->item);
  return c;
}

Genre *Genre_new(char *name) {
  Genre *g = malloc(sizeof(Genre));
  g->name = strdup(name);
  g->channels = NULL;
  g->item = gtk_menu_item_new_with_label(g->name);
  g->menu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(g->item), g->menu);
  gtk_widget_show(g->item);
  gtk_widget_show(g->menu);
  return g;
}

void Genre_append(Genre *g, Channel *c) {
  if (g && c) {
    g->channels = g_list_append(g->channels, c);
    gtk_menu_append(GTK_MENU(g->menu), c->item);
  }
}

/*
 * GET METHODS: get_info, get_spinner, get_channels, get_current
 */
int get_info(gchar **rurl, gchar **album, gchar **artist, gchar **song) {
  int len, count = 4;
  gchar *start, *end, *buf;
  *rurl = *album = *artist = *song = NULL;

  /* 
   * Having gotten the output from the web page earlier, see 
   * get_spinner(), we can now tokenize it on newlines and
   * iterate through the html searching for our artist/title/album
   * and rtsp url information
   */
  buf = (gchar *)strtok(data.raw, "\n");
  while (count > 0 && buf) {
    if (strstr(buf, "artist:")) {
      start = strstr(buf, "<span class=\"data\">");
      if (!start) break;
      start += 19;
      end = strstr(start, "</span>");
      if (!end) break;
      len = end - start;
      if (len < 2) { if (DEBUG) { printf("PRATIK 1\n"); } break; }
      *artist = malloc(len+1);
      strncpy(*artist, start, len);
      *((*artist)+len) = '\0';
      finish(*artist);
      count--;
    } else if (strstr(buf, "song:")) {
      start = strstr(buf, "<span class=\"data\">");
      if (!start) break;
      start += 19;
      end = strstr(start, "</span>");
      if (!end) break;
      len = end - start;
      if (len < 2) { if (DEBUG) { printf("PRATIK 2\n"); } break; }
      *song = malloc(len+1);
      strncpy(*song, start, len);
      *((*song)+len) = '\0';
      finish(*song);
      count--;
    } else if (strstr(buf, "album:")) {
      start = strstr(buf, "<span class=\"data\">");
      if (!start) break;
      start += 19;
      end = strstr(start, "</span>");
      if (!end) break;
      len = end - start;
      if (len < 2) { if (DEBUG) { printf("PRATIK: 3\n"); } break; }
      *album = malloc(len+1);
      strncpy(*album, start, len);
      *((*album)+len) = '\0';
      finish(*album);
      count--;
    } else if (strstr(buf, "embed")) {
      start = strstr(buf, "spinlinks");
      if (!start) break;
      start += 12;
      end = strstr(start, "%3F");
      if (!end) break;
      len = end - start;
      *rurl = malloc(len+1);
      if (len < 4) { 
	if (DEBUG) { printf("Hey Pratik - found the bug again:\n%s", buf); }
	break;
      }
      strncpy(*rurl, start, len);
      *((*rurl)+len) = '\0';
      count--;
    }
    buf = strtok(NULL, "\n");
  }
  free(data.raw);
  data.raw = NULL;
  data.pos = 0;
  data.length = 0;
  if (count) return 0;
  return 1;
}

size_t add_data(void *ptr, size_t size, size_t nmemb, void *stream) {
  size_t bytes = size*nmemb;
  int oldlen = data.length;
  while ((data.pos + bytes) > data.length) data.length += 4096;
  if (oldlen != data.length) data.raw = realloc(data.raw, data.length);
  strncpy(data.raw+data.pos, ptr, bytes);
  data.pos += bytes;
  return bytes;
}


void get_spinner(int seq, int id) {
  CURL *curl;
  char buf[512];
  snprintf(buf, 512, spinner_format, id);
  curl = curl_easy_init();
  curl_easy_setopt(curl, CURLOPT_URL, buf);
  curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, add_data);
  curl_easy_perform(curl);
  curl_easy_cleanup(curl);
  if (DEBUG) { printf("got spinner via url: %s\n",buf); }
}

GtkWidget *get_channels() {
  FILE *f;
  char buf[81] = "\0";
  Genre *currentGenre = NULL;
  GtkWidget *menu = gtk_menu_new();
  Genres = NULL;
  /* default channel */
  None = Channel_new("None", 0);
  gtk_menu_append(GTK_MENU(menu), None->item);
  /* open config file */
  if (DEBUG) { printf("Getting channels from file %s\n", CONFIG_PATH); }
  if ((f = fopen(CONFIG_PATH, "r"))) {
    /* read each genre, then add channels to current genre */
    while (fgets(buf, 80, f)) {
      if (*buf == 'g') {
	Genre *g;
	char name[65] = "\0";
	sscanf(buf, "genre %64c\n", name);
	name[strlen(name)-1] = '\0';
	g = Genre_new(name);
	gtk_menu_append(GTK_MENU(menu), g->item);
	Genres = g_list_append(Genres, g);
	currentGenre = g;
      } else if (*buf == 'c') {
	Channel *c;
	int id;
	char name[65] = "\0";
	sscanf(buf, "channel %d\t%64c\n", &id, name);
	name[strlen(name)-1] = '\0';
	c = Channel_new(name, id);
	Genre_append(currentGenre, c);
      }
    }
    fclose(f);
  }
  gtk_widget_show_all(menu);
  return menu;
}

Channel *get_current() {
  int id = 0;
  FILE *f;
  GList *gl, *gm = Genres;
  Channel *c;
  Genre *g;
  if ((f = fopen_home(".gspinner.channel", "r"))) {
    fscanf(f, "%d", &id);
    fclose(f);
  }
  if (id == 0) return None;
  while (gm) {
    /* traverse each genre */
    g = (Genre *)(gm->data);
    if (g) {
      gl = g->channels;
      while (gl) {
	/* traverse each Genre's channels */
	c = (Channel *)(gl->data);
	if (c && (c->id == id)) {
	  int index = g_list_index(Genres, g) + 1;
	  gtk_option_menu_set_history(GTK_OPTION_MENU(lmenu),
				      index);
	  return c;
	}
	gl = gl->next;
      }
    }
    gm = gm->next;
  }
  return None;
}

/*
 * MAIN METHODS: spawn, run_atexit, main
 */

void spawn() {
  int seq = 0;
  pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
  gchar *buf, *rurl, *album, *artist, *song;
  if (DEBUG) { printf("Child here within spawn\n"); }
  while (1) {
    if (current->id == 0) {
      if (DEBUG) { printf("current->id == 0\n"); }
      gtk_label_set_text(GTK_LABEL(lartist), "");
      gtk_label_set_text(GTK_LABEL(lalbum), "");
      gtk_label_set_text(GTK_LABEL(lsong), "");
      gtk_label_set_text(GTK_LABEL(lchannel), "");
      pthread_cond_wait(&cond, &mutex);
    }
    
    /* there's a periodic problem when the web page is
     * not returned or something similar occurs, let's
     * just make sure that get_info works right */
    { 
      int read_okay = -2;
      if (DEBUG) printf("=====>IN THE WHEEL\n");
      do { 
	if (DEBUG) printf("Spawn: calling get_spinner\n");
	get_spinner(++seq, current->id);
	if (DEBUG) printf("Spawn: calling get_info\n");
	read_okay = get_info(&rurl, &album, &artist, &song);
	if (DEBUG) printf("--->WHEEL TURN: %i\n", read_okay);
	if (read_okay == 0) { 
	  if (DEBUG) printf("--->FAILURE taking 5 before trying again\n");
	  sleep(5);
	}
      } while (read_okay != 1); 
    }
    
    if (DEBUG) { printf("setting labels now\n"); }
    /* set labels */
    
    gtk_label_set_text(GTK_LABEL(lartist), artist);
    if (DEBUG) { printf("artist = %s\n", artist); }

    gtk_label_set_text(GTK_LABEL(lalbum), album);
    if (DEBUG) {  printf("album = %s\n", album); }

    gtk_label_set_text(GTK_LABEL(lsong), song);
    if (DEBUG) { printf("song = %s\n", song); }

    gtk_label_set_text(GTK_LABEL(lchannel), current->name);
    if (DEBUG) { printf("channel = %s\n", current->name); }

    /* free memory */
    free(artist);
    free(album);
    free(song);
    buf = malloc(256);
    if (DEBUG) { printf("Calling trplayer\n"); }
    snprintf(buf, 255, trplayer_format, TRPLAYER, rurl);
    System(buf);
    free(buf);
    free(rurl);
  }
  pthread_exit(0);
}

void run_atexit() {
  FILE *f;
  if (trpid) kill(trpid, SIGKILL);
  if ((f = fopen_home(".gspinner.channel", "w"))) {
    fprintf(f, "%d", current->id);
    fclose(f);
  }
}

int main(int argc, char *argv[]) {
  char title[24];
  pthread_t thread;
  pthread_attr_t attr;
  GtkWidget *window, *table;
  /* usage */
  if (argc > 1) {
    int x;
    for (x = 1; x<argc; x++) {
      if (!strcmp(argv[x], "--version"))
	printf("gSpinner version %s\n", VERSION);
      if (!strcmp(argv[x], "--help"))
	printf("\nWho is John Galt?\n\n");
    }
    exit(0);
  }
  /* initialize some elements */
  gtk_init(&argc, &argv);
  atexit(run_atexit);
  setenv("REALPLAYER_HOME", REALPLAYER_DIR, 1);
  Genres = NULL;
  current = NULL;
  data.raw = NULL;
  data.pos = 0;
  data.length = 0;
  
  /* create the main window */
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  snprintf(title, 23, "gSpinner-%s", VERSION);
  gtk_window_set_title(GTK_WINDOW(window), title);
  gtk_window_set_default_size(GTK_WINDOW(window), 350, 100);
  gtk_window_set_policy(GTK_WINDOW(window), 0, 0, 0);
  gtk_container_set_border_width(GTK_CONTAINER(window), 5);
  gtk_signal_connect(GTK_OBJECT(window), "delete_event", gtk_main_quit, 
		     NULL);
  
  /* create the table to store labels and the menu */
  table = gtk_table_new(5, 2, FALSE);
  gtk_container_add(GTK_CONTAINER(window), table);
  gtk_table_set_row_spacings(GTK_TABLE(table), 0);
  gtk_table_set_col_spacings(GTK_TABLE(table), 5);
  
  /* create the options menu */
  lmenu = gtk_option_menu_new();
  gtk_widget_set_usize(lmenu, 150, 0);
  /* make new labels */
  lartist = gtk_label_new("");
  lalbum = gtk_label_new("");
  lsong = gtk_label_new("");
  lchannel = gtk_label_new("");
  
  /* pack table */
  gtk_table_attach(GTK_TABLE(table), gtk_label_new("Artist:"),
		   0, 1, 0, 1, GTK_SHRINK, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(table), gtk_label_new("Album:"),
		   0, 1, 1, 2, GTK_SHRINK, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(table), gtk_label_new("Song:"),
		   0, 1, 2, 3, GTK_SHRINK, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(table), gtk_label_new("Channel:"),
		   0, 1, 3, 4, GTK_SHRINK, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(table), lartist, 
		   1, 2, 0, 1, GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(table), lalbum, 
		   1, 2, 1, 2, GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(table), lsong, 
		   1, 2, 2, 3, GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(table), lchannel, 
		   1, 2, 3, 4, GTK_FILL, GTK_SHRINK, 0, 2);
  gtk_table_attach(GTK_TABLE(table), lmenu, 
		   0, 2, 4, 5, GTK_FILL, GTK_SHRINK, 0, 2);
  
  /* create the channel listings */
  gtk_option_menu_set_menu(GTK_OPTION_MENU(lmenu), get_channels());
  
  /* find current saved channel (if any) */
  current = get_current();
  
  /* make it all visible */
  gtk_widget_show_all(window);
  
  if (DEBUG) { printf("Spawning thread\n"); }
  /* thread junk */
  pthread_attr_init(&attr);
  if (pthread_create(&thread, &attr, (void *)&spawn, lmenu) < 0) {
    perror("pthread_create");
    exit(1);
  }
  if (DEBUG) { printf("Thread spawned - main going to gtk_main\n"); }
  gtk_main();
  return 0;
}
